package sem;

public class Semantics {

public static void SemanticAnalysis(ast.AST E) {
    if (E instanceof ast.PROGRAM) PROGRAM((ast.PROGRAM) E);
    else aux.Error.Internal("SemanticAnalysis", "Node " + E.getClass().getName() + " unexpected.");
}

public static void PROGRAM(ast.PROGRAM E) {
    E.currentProc = sym.Standard.MainProc = new sym.ProcedureSy("$MAIN", E.position, 0);

   E.decls.number = 0;
   E.decls.level = 0;
   E.decls.offsetIn = 0;

   E.decls.declsIn = new sym.SyTab();
   E.decls.env = sym.Standard.env();
   E.decls.currentProc = E.currentProc;
   Declarations(E.decls);

   sym.Standard.MainProc.SetProcFormals(new sym.SyTab());
   sym.Standard.MainProc.SetFormalSize(0);
   sym.SyTab locals = SemAux.ExtractVariables(E.decls.declsOut);
   sym.Standard.MainProc.SetProcLocals(locals);
   sym.Standard.MainProc.SetLocalSize(E.decls.offsetOut);

   E.stats.env = sym.Standard.env().cons(E.decls.declsOut);
   E.stats.isInLoop = false;
   E.stats.currentLoopExit = null;
   E.stats.currentProc = E.currentProc;
   SemAux.ResolveForwards(E.stats.env);
        
   E.stats.parentNext = E.label = new sym.Label();
   E.stats.offsetIn = E.decls.offsetOut;
   Statements(E.stats);
}

//===============================================================
//                         Declarations
//===============================================================

public static void Declarations(ast.AST E) {
   if (E instanceof ast.DECLNULL) DECLNULL((ast.DECLNULL)E);
   else if (E instanceof ast.VARDECL) VARDECL((ast.VARDECL)E);
   else if (E instanceof ast.CONSTDECL) CONSTDECL((ast.CONSTDECL)E);
   else if (E instanceof ast.REFDECL) REFDECL((ast.REFDECL)E);
   else if (E instanceof ast.ARRAYDECL) ARRAYDECL((ast.ARRAYDECL)E);
   else if (E instanceof ast.CLASSDECL) CLASSDECL((ast.CLASSDECL)E);
   else if (E instanceof ast.EXTENDSNULL) EXTENDSNULL((ast.EXTENDSNULL)E);
   else if (E instanceof ast.EXTENDS) EXTENDS((ast.EXTENDS)E);
   else if (E instanceof ast.CLASSFIELDDECL) CLASSFIELDDECL((ast.CLASSFIELDDECL)E);
   else if (E instanceof ast.METHODDECL) METHODDECL((ast.METHODDECL)E);
   else if (E instanceof ast.RECORDDECL) RECORDDECL((ast.RECORDDECL)E);
   else if (E instanceof ast.FIELDDECL) FIELDDECL((ast.FIELDDECL)E);
   else if (E instanceof ast.ENUMDECL) ENUMDECL((ast.ENUMDECL)E);
   else if (E instanceof ast.ENUMIDDECL) ENUMIDDECL((ast.ENUMIDDECL)E);
   else if (E instanceof ast.PROCDECL) PROCDECL((ast.PROCDECL)E);
   else if (E instanceof ast.FORMALDECL) FORMALDECL((ast.FORMALDECL)E);
   else if (E instanceof ast.DECLS) DECLS((ast.DECLS)E);
   else aux.Error.Internal("Declarations", "Node " + E.getClass().getName() + " unexpected.");
}

//---------------- List of Declarations ------------------------
public static void DECLS(ast.DECLS E) {
   if (E.declsIn.locateByName(E.left.ident) != null /*Standard.NoSy*/)
      aux.Error.SemId(E.position, "Multiple declaration", E.left.ident);
   
   E.left.number = E.number;
   E.left.offsetIn = E.offsetIn;
   E.left.level = E.level;
   E.left.currentSy = E.right.currentSy = E.currentSy;

   E.left.env = E.env.cons(E.declsIn);
   E.left.currentProc = E.currentProc;
   Declarations(E.left);

   E.right.number = E.number + 1;
   E.right.offsetIn = E.left.offsetOut;
   E.right.level = E.level;

   E.right.declsIn = E.declsIn.insert(E.left.symbol);
   E.right.env = E.env;
   E.right.currentProc = E.currentProc;
   Declarations(E.right);
   E.declsOut = E.right.declsOut;

   E.offsetOut = E.right.offsetOut;
}

public static void DECLNULL(ast.DECLNULL E) {
   E.declsOut = E.declsIn;
   E.offsetOut = E.offsetIn;
}


//---------------- Variable Declarations ------------------------
public static void VARDECL(ast.VARDECL E) {
   sym.TypeSy Type = SemAux.ExpectType(E.typeName, E.env, E.position);
   sym.VariableSy v = new sym.VariableSy(E.ident, E.position, E.level);
   E.symbol = v;

   if (Type != null) {
      v.SetType(Type);
      E.offsetOut = sym.Arch.AddAlign(E.offsetIn, Type);
      v.SetOffset(sym.Arch.CalcOffset(E.offsetIn, Type));
      v.SetSize(Type.GetSize());
   }
}

//---------------- Constant Declarations ------------------------
public static void CONSTDECL(ast.CONSTDECL E) {
   sym.TypeSy Type = SemAux.ExpectType(E.typeName, E.env, E.position);
   Type = SemAux.CheckScalar(Type, E.expr.position);
   sym.ConstSy c = new sym.ConstSy(E.ident, E.position, E.level);
   E.symbol = c;

   E.expr.env = E.env;
   E.expr.currentProc = E.currentProc;
   Expression(E.expr);

   if (E.expr.type != Type &&
      Type != null /*Standard.NoType*/ &&
      E.expr.type != null /*Standard.NoType*/)
      aux.Error.Sem(E.position, "Wrong expression type");
   if (!E.expr.isConst && E.expr.type != null /*Standard.NoType*/)
      aux.Error.Sem(E.position, "Constant expression expected");

   if (Type != null) {
      c.SetConstantValue(E.expr.value);
      c.SetType(Type);
      c.SetSize(Type.GetSize());
   }

   E.offsetOut = E.offsetIn;
}

//---------------- Ref Declarations ------------------------
public static void REFDECL(ast.REFDECL E) {
   sym.RefType t = new sym.RefType(E.ident, E.position, E.level);
   E.symbol = t;
   SemAux.AddForwardRef(t, E.typeName, E.position);
   t.SetSize(sym.Standard.AddrType.GetSize());
   E.offsetOut = E.offsetIn;
}

//---------------- Array Declarations ------------------------
public static void ARRAYDECL(ast.ARRAYDECL E) {
   sym.TypeSy Type = SemAux.ExpectType(E.elementTypeName, E.env, E.position);
   sym.ArrayType t = new sym.ArrayType(E.ident, E.position, E.level);
   E.symbol = t;

   E.count.env = E.env;
   E.count.currentProc = E.currentProc;
   Expression(E.count);

   if (E.count.type != sym.Standard.IntType &&
      E.count.type != null /*Standard.NoType*/)
      aux.Error.Sem(E.position, "Integer Expression expected");
   if (!E.count.isConst)
      aux.Error.Sem(E.position, "Constant Expression expected");

   if (Type != null) {
      int v = 0;
      if (E.count.value != null && E.count.value instanceof sym.IntValue)
          v = ((sym.IntValue)E.count.value).getValue();

      t.SetArrayCount(v);
      t.SetArrayElementType(Type);
      t.SetSize(v * sym.Arch.CalcOffset(
                      Type.GetSize(),
                      SemAux.GetLargestElementType(Type)));
   }

   E.offsetOut = E.offsetIn;
}

//---------------- Class Declarations ------------------------
public static void CLASSDECL(ast.CLASSDECL E) {
   if (E.level != 0) {
      aux.Error.SemId(E.position, "Classes may only be declared at the global level", E.ident);
   }
   E.extends_.env = E.env;
   E.extends_.currentProc = E.currentProc;
   Declarations(E.extends_);

   sym.ClassType t = new sym.ClassType(E.ident, E.position, E.level);
   E.symbol = t;
   sym.ClassType superType = (sym.ClassType)E.extends_.symbol;
   t.SetSuperType(superType);
   t.SetSize(sym.Standard.AddrType.GetSize());

   E.fields.declsIn = new sym.SyTab();
   E.fields.env = E.env.cons(new sym.SyTab(t));
   E.fields.number = 0;
   E.fields.currentSy = t;
   E.fields.offsetIn = superType.GetFieldSize();
   E.fields.currentProc = E.currentProc;
   Declarations(E.fields);
   t.SetFields(E.fields.declsOut);
   t.SetFieldSize(E.fields.offsetOut);

   E.methods.declsIn = E.fields.declsOut;
   E.methods.env = E.fields.env;
   E.methods.number = 0;
   E.methods.currentSy = t;
   E.methods.offsetIn = superType.GetMethodSize();
   E.methods.currentProc = E.currentProc;
   Declarations(E.methods);
   t.SetFields(E.methods.declsOut);
   t.SetMethodSize(E.methods.offsetOut);

   E.offsetOut = E.offsetIn;
}

public static void EXTENDS(ast.EXTENDS E) {
   sym.Symbol t = SemAux.ExpectType(E.ident, E.env, E.position);
   E.symbol = t;
   if ((t != null /*Standard.NoType*/) &&
       (!(E.symbol instanceof sym.ClassType))) {
      aux.Error.SemId(E.position, "Class type expected", E.ident);
      E.symbol = sym.Standard.RootSy;
   }
}

public static void EXTENDSNULL(ast.EXTENDSNULL E) {
   E.symbol = sym.Standard.RootSy;
}

public static void CLASSFIELDDECL(ast.CLASSFIELDDECL E) {
   sym.TypeSy ty = SemAux.ExpectType (E.typeName, E.env, E.position);
   sym.FieldSy f = new sym.FieldSy(E.ident, E.position, E.level);
   sym.ClassType cl = (sym.ClassType) E.currentSy;
   E.symbol = f;

   if (ty != null) {
      f.SetType(ty);
      f.SetParent(cl);
      E.offsetOut = sym.Arch.AddAlign(E.offsetIn, ty);
      f.SetOffset(sym.Arch.CalcOffset(E.offsetIn, ty));
      f.SetSize(ty.GetSize());
   }
}

//---------------- Method Declarations ------------------------
public static void METHODDECL(ast.METHODDECL E) {
   sym.MethodSy m = new sym.MethodSy(E.ident, E.position, E.level);
   E.symbol = m;
   sym.ClassType cl = (sym.ClassType) E.currentSy;
   m.SetParent(cl);
   cl.SetFields(cl.GetFields().insert(m));

   sym.ProcedureSy Proc = new sym.ProcedureSy(E.ident, E.position, E.level);
   m.SetMethodProc(Proc);

   // Construct SELF;
   sym.SyTab Temp = new sym.SyTab();
   sym.FormalSy Self = new sym.FormalSy("SELF", E.position, E.level+1);
   Self.SetFormalNumber(1);
   Self.SetFormalMode("VAL");
   Self.SetType(cl);
   Self.SetOffset(0);
   Self.SetSize(sym.Standard.AddrType.GetSize());
   Temp = Temp.insert(Self);

   // Process formals;
   E.formals.declsIn = Temp;
   E.formals.env = E.env;
   E.formals.number = 1;
   E.formals.offsetIn = sym.Standard.AddrType.GetSize();
   E.formals.level = E.level+1;
   E.formals.currentProc = Proc; // m;
   Declarations(E.formals);
   Proc.SetProcFormals(E.formals.declsOut);
   Proc.SetFormalSize(E.formals.offsetOut);

   // Process locals;
   E.decls.declsIn = E.formals.declsOut.clone();
   E.decls.env = E.env;
   E.decls.number = 0;
   E.decls.offsetIn = 0;
   E.decls.level = E.level + 1;
   E.decls.currentProc = Proc; // m;
   Declarations(E.decls);
   SemAux.ResolveForwards(E.env.cons(E.decls.declsOut));
   Proc.SetProcLocals(SemAux.ExtractVariables(E.decls.declsOut));
   Proc.SetLocalSize(E.decls.offsetOut);

   // Process Statements;
   sym.Env Temp1 = SemAux.ConcatClassFields(cl);
   Temp1.append(E.env);
   E.stats.env = Temp1.cons(new sym.SyTab(m)).cons(E.decls.declsOut);
   E.stats.parentNext = E.label = new sym.Label();
   E.stats.isInLoop = false;
   E.stats.currentLoopExit = null;
   E.stats.currentProc = Proc; //m;
   E.stats.offsetIn = E.decls.offsetOut;
   Statements(E.stats);

   E.offsetOut = E.offsetIn;

   sym.ClassType su = (sym.ClassType) cl.GetSuperType();
   sym.Symbol OverriddenMethod = SemAux.LocateClassField(su, E.ident);
   if (OverriddenMethod == null /*Standard.NoSy*/) {
      m.SetOffset(E.offsetIn);
      E.offsetOut = E.offsetIn + sym.Standard.AddrType.GetSize();
   } else {
       if (! (OverriddenMethod instanceof sym.MethodSy))
          aux.Error.SemId(E.position, "Only methods can be overridden", E.ident);
      else {
         sym.MethodSy orm = (sym.MethodSy) OverriddenMethod;
         m.SetOffset(orm.GetOffset());
         if (!SemAux.CheckSignature(m.GetMethodProc(),orm.GetMethodProc()))
             aux.Error.SemId(E.position, "Type error in method overriding", E.ident);
      }
      E.offsetOut = E.offsetIn;
   }
}

//---------------- Record Declarations ------------------------
public static void RECORDDECL(ast.RECORDDECL E) {
   sym.RecordType t = new sym.RecordType(E.ident, E.position, E.level);
   E.symbol = t;

   E.fields.declsIn = new sym.SyTab();
   E.fields.env = E.env;
   E.fields.number = 0;
   E.fields.offsetIn = 0;
   E.fields.currentSy = E.symbol;
   E.fields.currentProc = E.currentProc;
   Declarations(E.fields);

   t.SetFields(E.fields.declsOut);
   t.SetSize(E.fields.offsetOut);
   E.offsetOut = E.offsetIn;
}

public static void FIELDDECL(ast.FIELDDECL E) {
   sym.TypeSy Type = SemAux.ExpectType(E.typeName, E.env, E.position);
   sym.FieldSy t = new sym.FieldSy(E.ident, E.position, E.level);
   sym.RecordType re = (sym.RecordType) E.currentSy;
   E.symbol = t;
   if (Type != null) {
      t.SetType(Type);
      t.SetParent(re);
      E.offsetOut = sym.Arch.AddAlign(E.offsetIn, Type);
      t.SetOffset(sym.Arch.CalcOffset(E.offsetIn, Type));
      t.SetSize(Type.GetSize());
   }

}

//---------------- Enumeration Declarations ------------------------
public static void ENUMDECL(ast.ENUMDECL E) {
   E.enumIds.declsIn = new sym.SyTab();
   E.enumIds.env = E.env;
   E.enumIds.number = 0;
   sym.EnumType t = new sym.EnumType(E.ident, E.position, E.level);
   E.enumIds.currentSy = E.symbol = t;
   E.enumIds.currentProc = E.currentProc;
   Declarations(E.enumIds);
   t.SetEnumElements(E.enumIds.declsOut);
   t.SetSize(sym.Standard.IntType.GetSize());
   E.offsetOut = E.offsetIn;
}

public static void ENUMIDDECL(ast.ENUMIDDECL E) {
    sym.EnumSy t = new sym.EnumSy(E.ident, E.position, E.level);
    E.symbol = t;
    sym.EnumType en = (sym.EnumType) E.currentSy;

   E.expr.currentProc = E.currentProc;
   E.expr.env = E.env;
   Expression(E.expr);
   if (!E.expr.isConst)
      aux.Error.Sem(E.position, "Constant Expression expected");
   if (E.expr.type != sym.Standard.IntType)
      aux.Error.Sem(E.position, "Integer Expression expected");
   if (E.expr.value != null && E.expr.value instanceof sym.IntValue)
       t.SetEnumValue(((sym.IntValue)E.expr.value).getValue());
         
   t.SetType(en);
   t.SetSize(sym.Standard.IntType.GetSize());
}

//---------------- Procedure Declarations ------------------------
public static void PROCDECL(ast.PROCDECL E) {
   sym.ProcedureSy t = new sym.ProcedureSy(E.ident, E.position, E.level);
   E.symbol = t;

   E.formals.declsIn = new sym.SyTab();
   E.formals.env = E.env;
   E.formals.number = 0;
   E.formals.offsetIn = 0;
   E.formals.level = E.level + 1;
   E.formals.currentProc = t;
   Declarations(E.formals);
   t.SetProcFormals(E.formals.declsOut);
   t.SetFormalSize(E.formals.offsetOut);

   SemAux.LargeValueFormalCopies(t, E.formals.declsOut, E.position);
   sym.SyTab copies = t.GetProcLocals();

   E.decls.declsIn = E.formals.declsOut.clone().merge(copies);
   E.decls.env = E.env.cons(new sym.SyTab(t));
   E.decls.number = copies.count();
   E.decls.offsetIn = t.GetLocalSize();
   E.decls.level = E.level + 1;
   E.decls.currentProc = t;
   Declarations(E.decls);
   SemAux.ResolveForwards(E.env.cons(E.decls.declsOut));
   t.SetProcLocals(SemAux.ExtractVariables(E.decls.declsOut));
   t.SetLocalSize(E.decls.offsetOut);

   E.stats.env = E.env.cons(new sym.SyTab(t)).cons(E.decls.declsOut);
   E.stats.parentNext = E.label = new sym.Label();
   E.stats.isInLoop = false;
   E.stats.currentLoopExit = null;
   E.stats.currentProc = t;
   E.stats.offsetIn = E.decls.offsetOut;
   Statements(E.stats);

   E.offsetOut = E.offsetIn;
}

public static void FORMALDECL(ast.FORMALDECL E) {
   sym.TypeSy Type = SemAux.ExpectType(E.typeName, E.env, E.position);
   sym.FormalSy t = new sym.FormalSy(E.ident, E.position, E.level);
   E.symbol = t;
   t.SetFormalNumber(E.number+1);
   t.SetFormalMode(E.mode);

   if (Type != null) {
      t.SetType(Type);
      sym.TypeSy Temp = null;
      if (E.mode.equals("VAR"))
         Temp = sym.Standard.AddrType;
      else if (Type.GetSize() > sym.Standard.IntType.GetSize())
         Temp = sym.Standard.AddrType;
      else if (Type.GetSize() < sym.Standard.IntType.GetSize())
         Temp = sym.Standard.IntType;
      else
         Temp = Type;
      E.offsetOut = sym.Arch.AddAlign(E.offsetIn, Temp);
      t.SetOffset( sym.Arch.CalcOffset(E.offsetIn, Temp));
      t.SetSize(Type.GetSize());
   }
}

//===============================================================
//                         Statements
//===============================================================

public static void Statements(ast.AST E) {
   if (E instanceof ast.STATNULL) STATNULL((ast.STATNULL)E);
   else if (E instanceof ast.ASSIGN) ASSIGN((ast.ASSIGN)E);
   else if (E instanceof ast.BLOCK) BLOCK((ast.BLOCK)E);
   else if (E instanceof ast.PROCCALL) PROCCALL((ast.PROCCALL)E);
   else if (E instanceof ast.WRITE) WRITE((ast.WRITE)E);
   else if (E instanceof ast.SPECIAL) SPECIAL((ast.SPECIAL)E);
   else if (E instanceof ast.WRITELN) WRITELN((ast.WRITELN)E);
   else if (E instanceof ast.READ) READ((ast.READ)E);
   else if (E instanceof ast.WHILE) WHILE((ast.WHILE)E);
   else if (E instanceof ast.REPEAT) REPEAT((ast.REPEAT)E);
   else if (E instanceof ast.LOOP) LOOP((ast.LOOP)E);
   else if (E instanceof ast.EXIT) EXIT((ast.EXIT)E);
   else if (E instanceof ast.IF1) IF1((ast.IF1)E);
   else if (E instanceof ast.IF2) IF2((ast.IF2)E);
   else if (E instanceof ast.FOR) FOR((ast.FOR)E);
   else if (E instanceof ast.STATS) STATS((ast.STATS)E);
   else aux.Error.Internal("Statements", "Node " +   E.getClass().getName() + " unexpected.");
}


//---------------- List of Statements ------------------------
public static void STATS(ast.STATS E) {
   E.left.env = E.right.env = E.env;
   E.left.isInLoop = E.right.isInLoop = E.isInLoop;
   E.left.currentLoopExit = E.right.currentLoopExit = E.currentLoopExit;
   if (E.right instanceof ast.STATNULL)
      E.left.next = E.parentNext;
   else
      E.left.next = E.right.label;
   E.left.currentProc = E.currentProc;
   E.left.offsetIn = E.offsetIn;
   E.left.declsIn = E.declsIn;
   E.left.label = new sym.Label();
   Statements(E.left);

   E.right.parentNext = E.parentNext;
   E.right.currentProc = E.currentProc;
   E.right.offsetIn = E.offsetIn;
   E.right.declsIn = E.left.declsOut;
   Statements(E.right);
   E.offsetOut = (E.left.offsetOut > E.right.offsetOut)?E.left.offsetOut:E.right.offsetOut;
   E.declsOut = E.right.declsOut;
}

public static void STATNULL(ast.STATNULL E) {
   E.offsetOut = E.offsetIn;
   E.declsOut = E.declsIn;
}

//---------------- Assignment Statement ------------------------
public static void ASSIGN(ast.ASSIGN E) {
   E.left.env = E.right.env = E.env;
   E.left.currentProc = E.currentProc;
   Expression(E.left);
   E.right.true_ = new sym.Label();
   E.right.false_ = new sym.Label();
   E.right.currentProc = E.currentProc;
   Expression(E.right);
   E.left.type  = SemAux.CheckAssignable(E.left.type, E.position);
   if (!SemAux.Assignable(E.left.type, E.right.type) &&
      E.left.type != null /*Standard.NoType*/ &&
       E.right.type != null /*Standard.NoType*/)
       aux.Error.Sem(E.position, "Type missmatch in assignment statement.");
   if (!E.left.isLValue && E.left.type != null)
       aux.Error.Sem(E.position, "Can't assign to a constant or expression.");

   E.offsetOut = E.offsetIn;
   E.declsOut = E.declsIn;
}

    //---------------- Block Statement ------------------------
 public static void BLOCK(ast.BLOCK E) {
     E.decls.number = 0;
     E.decls.level = E.currentProc.GetLevel();
     E.decls.offsetIn = E.offsetIn;
     E.decls.declsIn = E.declsIn;
     E.decls.env = E.env;
     E.decls.currentProc = E.currentProc;
     Declarations(E.decls);

     E.stats.env = E.env.cons(E.decls.declsOut);
     E.stats.isInLoop = false;
     E.stats.currentLoopExit = E.currentLoopExit;
     E.stats.currentProc = E.currentProc;
     E.stats.offsetIn = E.decls.offsetOut;
     E.stats.declsIn =  E.decls.declsOut;
     Statements(E.stats);
     E.offsetOut = E.stats.offsetOut;
     E.declsOut = E.stats.declsOut;
}

//--------------------- Procedure Calls  ------------------------
public static void PROCCALL(ast.PROCCALL E) {
    E.des.env = E.env;
    E.des.currentProc = E.currentProc;
    Expression(E.des);
    E.symbol = E.des.symbolOut;

    if (E.symbol != null)
	if (!(E.symbol instanceof sym.ProcedureSy) && !(E.symbol instanceof sym.MethodSy))
	    aux.Error.Sem(E.position, "Procedure identifier expected");
        else {
	    if (E.symbol instanceof sym.ProcedureSy) {
		sym.ProcedureSy pr = (sym.ProcedureSy) E.symbol;
		E.actuals.formals = pr.GetProcFormals();
                E.actuals.formalNumber = 1;
            } else {
		sym.MethodSy me = (sym.MethodSy) E.symbol;
                sym.ProcedureSy pr = me.GetMethodProc();
                E.actuals.formals = pr.GetProcFormals().deleteByName("SELF");
                E.actuals.formalNumber = 2;
	    }
            E.actuals.env = E.env;
            E.actuals.currentProc = E.currentProc;
            ActualParams(E.actuals);
         }
    E.offsetOut = E.offsetIn;
    E.declsOut = E.declsIn;
}

//--------------------- Read/Write Statements ---------------------
public static void WRITE(ast.WRITE E) {
   E.expr.env = E.env;
   E.expr.currentProc = E.currentProc;
   Expression(E.expr);
   SemAux.CheckWriteable(E.expr.type, E.expr.position);
   E.offsetOut = E.offsetIn;
   E.declsOut = E.declsIn;
}

public static void SPECIAL(ast.SPECIAL E) {
   E.offsetOut = E.offsetIn;
   E.declsOut = E.declsIn;
}

public static void WRITELN(ast.WRITELN E) {
   E.offsetOut = E.offsetIn;
   E.declsOut = E.declsIn;
}

public static void READ(ast.READ E) {
   E.des.env = E.env;
   E.des.currentProc = E.currentProc;
   Designator(E.des);
   SemAux.CheckReadable(E.des.type, E.des.position);
   if ((!E.des.isLValue) && E.des.type != null /*Standard.NoType*/)
      aux.Error.Sem(E.position, "Can't read to a constant.");
   E.offsetOut = E.offsetIn;
   E.declsOut = E.declsIn;
}

//------------------------- Control Statements ---------------------
public static void WHILE(ast.WHILE E) {
   E.expr.env = E.stats.env = E.env;
   E.begin = E.label;
   E.expr.true_ = new sym.Label();
   E.expr.false_ = E.next;
   E.expr.currentProc = E.currentProc;
   Expression(E.expr);
   if (E.expr.type != sym.Standard.BoolType && E.expr.type != null /*Standard.NoType*/)
      aux.Error.Sem(E.expr.position, "Boolean type expected ");
   E.stats.parentNext = E.begin;
   E.stats.isInLoop = E.isInLoop;
   E.stats.currentLoopExit = E.currentLoopExit;
   E.stats.currentProc = E.currentProc;
   E.stats.offsetIn = E.offsetIn;
   E.stats.declsIn = E.declsIn;
   Statements(E.stats);
   E.offsetOut = E.offsetIn;
   E.declsOut = E.stats.declsOut;
}

public static void REPEAT(ast.REPEAT E) {
   E.expr.env = E.stats.env = E.env;
   E.expr.false_ = E.label;
   E.expr.true_ = E.next;
   E.begin = new sym.Label();   // Start of code for expression;
   E.expr.currentProc = E.currentProc;
   Expression(E.expr);
   if (E.expr.type != sym.Standard.BoolType && E.expr.type != null /*Standard.NoType*/)
      aux.Error.Sem(E.expr.position, "Boolean type expected ");
   E.stats.next = E.stats.parentNext = E.begin;
   E.stats.isInLoop = E.isInLoop;
   E.stats.currentLoopExit = E.currentLoopExit;
   E.stats.currentProc = E.currentProc;
   E.stats.offsetIn = E.offsetIn;
   E.stats.declsIn = E.declsIn;
   Statements(E.stats);
   E.offsetOut = E.offsetIn;
   E.declsOut = E.stats.declsOut;
}

public static void LOOP(ast.LOOP E) {
   E.stats.env = E.env;
   E.begin = E.label;
   E.stats.parentNext = E.begin;
   E.stats.isInLoop = true;
   E.stats.currentLoopExit = E.next;
   E.stats.currentProc = E.currentProc;
   E.stats.offsetIn = E.offsetIn;
   E.stats.declsIn = E.declsIn;
   Statements(E.stats);
   E.offsetOut = E.offsetIn;
   E.declsOut = E.stats.declsOut;
}

public static void EXIT(ast.EXIT E) {
   if (!E.isInLoop)
      aux.Error.Sem(E.position, "EXIT only within LOOP ");
   E.offsetOut = E.offsetIn;
   E.declsOut = E.declsIn;
}

public static void IF1(ast.IF1 E) {
   E.expr.env = E.then_.env = E.env;
   E.expr.true_ = new sym.Label();
   E.expr.false_ = E.next;
   E.expr.currentProc = E.currentProc;
   Expression(E.expr);
   if (E.expr.type != sym.Standard.BoolType && E.expr.type != null /*Standard.NoType*/)
      aux.Error.Sem(E.expr.position, "Boolean type expected ");
   E.then_.parentNext = E.next;
   E.then_.isInLoop = E.isInLoop;
   E.then_.currentLoopExit = E.currentLoopExit;
   E.then_.currentProc = E.currentProc;
   E.then_.offsetIn = E.offsetIn;
   E.then_.declsIn = E.declsIn;
   Statements(E.then_);
   E.offsetOut = E.offsetIn;
   E.declsOut = E.then_.declsOut;
}

public static void IF2(ast.IF2 E) {
   E.expr.env = E.then_.env = E.else_.env = E.env;
   E.expr.true_ = new sym.Label();
   E.expr.false_ = new sym.Label();
   E.expr.currentProc = E.currentProc;
   Expression(E.expr);
   if (E.expr.type != sym.Standard.BoolType && E.expr.type != null /*Standard.NoType*/)
      aux.Error.Sem(E.expr.position, "Boolean type expected ");
   E.then_.next = E.else_.next = E.next;
   E.then_.parentNext = E.else_.parentNext = E.next;
   E.then_.isInLoop =  E.else_.isInLoop = E.isInLoop;
   E.then_.currentLoopExit =  E.else_.currentLoopExit = E.currentLoopExit;
   E.then_.currentProc = E.else_.currentProc = E.currentProc;
   E.then_.offsetIn = E.else_.offsetIn = E.offsetIn;
   E.then_.declsIn = E.declsIn;
   Statements(E.then_);
   E.else_.declsIn = E.then_.declsOut;
   Statements(E.else_);
   E.offsetOut = (E.else_.offsetOut > E.then_.offsetOut)?E.else_.offsetOut:E.then_.offsetOut;
   E.declsOut = E.else_.declsOut;
}

public static void FOR(ast.FOR E) {
   E.symbol = E.env.locateByName(E.idx);
   if (E.symbol == null /*Standard.NoSy*/)
       aux.Error.SemId(E.position, "FOR index variable not declared", E.idx);
   else {
      if (!(E.symbol instanceof sym.VariableSy))
          aux.Error.SemId(E.position, "Variable expected", E.idx);
      else {
         sym.VariableSy v = (sym.VariableSy) E.symbol;
         sym.TypeSy Type = v.GetType();

         if (E.by == null)
            if (Type == sym.Standard.IntType)
               E.by = new ast.INTLIT("1", E.to.position);
            else if (Type == sym.Standard.RealType)
               E.by = new ast.REALLIT("1.0", E.to.position);
            else if (Type == sym.Standard.CharType)
               E.by = new ast.INTLIT("1", E.to.position);
         E.from.env = E.to.env = E.by.env = E.stats.env = E.env;
         E.from.currentProc = E.to.currentProc = E.by.currentProc = E.currentProc;
         Expression(E.from);
         Expression(E.to);
         Expression(E.by);
         if (Type != E.from.type)
            aux.Error.Sem(E.from.position, "Type missmatch in FOR-statement FROM Expression");
         if (Type != E.to.type)
            aux.Error.Sem(E.to.position, "Type missmatch in FOR-statement TO Expression");
         if (Type != E.by.type)
            aux.Error.Sem(E.by.position, "Type missmatch in FOR-statement BY Expression");
         if (!E.by.isConst)
            aux.Error.Sem(E.by.position, "BY Expression must be constant");
         else
            E.byValue = E.by.value;
         if (E.currentProc == sym.Standard.MainProc)
            E.tempTo = SemAux.MakeTempVar(sym.Standard.MainProc, Type, 0, E.position);
         else
            E.tempTo = SemAux.MakeTempVar(
                           E.currentProc, Type,
                           E.currentProc.GetLevel()+1, E.position);

         E.forLabel = new sym.Label();
         E.endLabel = new sym.Label();
         E.incrLabel = new sym.Label();
         E.stats.next = E.stats.parentNext = E.incrLabel;
         E.stats.isInLoop = E.isInLoop;
         E.stats.currentLoopExit  = E.currentLoopExit;
         E.stats.currentProc = E.currentProc;
         E.stats.offsetIn = E.offsetIn;
         E.stats.declsIn = E.declsIn;
         Statements(E.stats);
      }
   }
   E.offsetOut = E.offsetIn;
   E.declsOut = E.stats.declsOut;
}


//===============================================================
//                 Procedure actual parameters
//===============================================================
public static void ActualParams(ast.ACTUAL E) {
   if (E instanceof ast.ACTUALNULL) ACTUALNULL((ast.ACTUALNULL)E);
   else if (E instanceof ast.ACTUAL) ACTUAL((ast.ACTUAL)E);
   else aux.Error.Internal("ActualParams", "Node " +   E.getClass().getName() + " unexpected.");
}

public static void ACTUAL(ast.ACTUAL E) {
   E.expr.env = E.nextActual.env = E.env;
   E.expr.currentProc = E.currentProc;
   Expression(E.expr);
   E.formal = sym.ProcedureSy.GetFormalParam(E.formals,E.formalNumber);
   if ((E.formal == null /*Standard.NoSy*/))
       aux.Error.Sem(E.position, "Too many actual parameters.");
   else {
      sym.TypeSy Type = E.formal.GetType();
      if (!SemAux.Assignable(Type,E.expr.type) &&
          Type != null /*Standard.NoType*/ &&
          E.expr.type != null /*Standard.NoType*/)
          aux.Error.Sem(E.position, "Actual/formal parameter type missmatch.");
      if ((!E.expr.isLValue) && (E.formal.GetFormalMode().equals("VAR")))
          aux.Error.Sem(E.position, "VAR formal parameter requires variable actual.");
      E.nextActual.formals = E.formals;
      E.nextActual.formalNumber = E.formalNumber + 1;
      E.nextActual.currentProc = E.currentProc;
      ActualParams(E.nextActual);
   }
}

public static void ACTUALNULL(ast.ACTUALNULL E) {
    sym.FormalSy Temp = sym.ProcedureSy.GetFormalParam(E.formals, E.formalNumber);
   if ((Temp != null /*Standard.NoSy*/))
       aux.Error.Sem(E.position, "Too few actual parameters.");
}


//===============================================================
//                         Designators
//===============================================================
public static void Designator(ast.DESIGNATOR E) {
    if (E instanceof ast.VARREF) VARREF((ast.VARREF)E);
    else if (E instanceof ast.DESNULL) DESNULL((ast.DESNULL)E);
    else if (E instanceof ast.INDEX) INDEX((ast.INDEX)E);
    else if (E instanceof ast.DEREF) DEREF((ast.DEREF)E);
    else if (E instanceof ast.FIELDREF) FIELDREF((ast.FIELDREF)E);
    else if (E instanceof ast.NARROW) NARROW((ast.NARROW)E);
    else if (E instanceof ast.CLASSREF) CLASSREF((ast.CLASSREF)E);
    else if (E instanceof ast.ENUMREF) ENUMREF((ast.ENUMREF)E);
    else aux.Error.Internal("Designator", "Node " +   E.getClass().getName() + " unexpected.");
}


public static void VARREF(ast.VARREF E) {
   E.type = null /*Standard.NoType*/;
   E.symbol = null /*Standard.NoSy*/;
   E.isLValue = true;
   E.isConst = false;
   E.next.typeOut = null /*Standard.NoType*/;
   E.next.symbolOut = null /*Standard.NoSy*/;
   E.symbolOut = null /*Standard.NoSy*/;

   E.symbol = E.env.locateByName(E.ident);
   if (E.symbol == null /*Standard.NoSy*/) {
       aux.Error.SemId(E.position, "Identifier not declared", E.ident);
   } else {
       if (E.symbol == sym.Standard.TrueSy) {
          E.isLValue = false;
          E.isConst = true;
          E.value = new sym.BoolValue(true);
      } else if (E.symbol == sym.Standard.FalseSy) {
          E.isLValue = false;
          E.isConst = true;
          E.value = new sym.BoolValue(false);
      } else if (E.symbol instanceof sym.ConstSy ) {
          sym.ConstSy co = (sym.ConstSy) E.symbol;
          E.value = co.GetConstantValue();
          E.isLValue = false;
          E.isConst = true;
      } else if (E.symbol instanceof sym.EnumSy) {
          sym.EnumSy en = (sym.EnumSy) E.symbol;
          E.value = new sym.IntValue(en.GetEnumValue());
          E.isLValue = false;
          E.isConst = true;
      } else if (E.symbol instanceof sym.VariableSy ||
                 E.symbol instanceof sym.FormalSy ||
                 E.symbol instanceof sym.FieldSy) {
          E.isLValue = true;
          E.isConst = false;
      }

      if (! ((E.symbol instanceof sym.VariableSy ||
             E.symbol instanceof sym.ProcedureSy ||
	      E.symbol instanceof sym.MethodSy))) {
          E.type = null /*Standard.NoType*/;
          aux.Error.SemId(E.position, "Variable expected", E.ident);
      } else {
         if (E.symbol instanceof sym.ProcedureSy ||
             E.symbol instanceof sym.MethodSy) 
            E.next.typeIn = null /*Standard.NoType*/;
         else {
            sym.VariableSy va = (sym.VariableSy) E.symbol;
            E.next.typeIn = va.GetType();
	 }
         E.next.symbolIn = E.symbol;
         E.next.env = E.env;
         E.next.currentProc = E.currentProc;
         Designator(E.next);
         E.type = E.next.typeOut;
         E.symbolOut = E.next.symbolOut;
     }
   }
}

public static void DESNULL(ast.DESNULL E) {
   E.typeOut = E.typeIn;
   E.symbolOut = E.symbolIn;
}

public static void INDEX(ast.INDEX E) {
   E.type = null /*Standard.NoType*/;
   E.next.typeOut = null /*Standard.NoType*/;
   E.next.symbolOut = null /*Standard.NoSy*/;

   E.index.env = E.env;
   E.index.currentProc = E.currentProc;
   Expression(E.index);
   E.index.type = SemAux.CheckInteger(E.index.type, E.position);

   if (! (E.typeIn instanceof sym.ArrayType)) {
       aux.Error.Sem(E.position, "Array variable expected");
       E.type = E.typeOut = null /*Standard.NoType*/;
   } else {
      E.next.env = E.env;
      sym.ArrayType t = (sym.ArrayType) E.typeIn;
      E.next.typeIn = t.GetArrayElementType();
      E.next.symbolIn = null /*Standard.NoSy*/;
      E.next.currentProc = E.currentProc;
      Designator(E.next);
      E.type = E.typeOut = E.next.typeOut;
      E.symbolOut = E.next.symbolOut;
   }
}

public static void DEREF(ast.DEREF E) {
   E.type = null /*Standard.NoType*/;
   E.next.typeOut = null /*Standard.NoType*/;
   E.next.symbolOut = null /*Standard.NoSy*/;
   if (! (E.typeIn instanceof sym.RefType)) {
       aux.Error.Sem(E.position, "Reference variable expected");
       E.type = E.typeOut = null /*Standard.NoType*/;
   } else {
      E.next.env = E.env;
      sym.RefType t = (sym.RefType) E.typeIn;
      E.next.typeIn = t.GetRefType();
      E.next.symbolIn = null /*Standard.NoSy*/;
      E.next.currentProc = E.currentProc;
      Designator(E.next);
      E.type = E.typeOut = E.next.typeOut;
      E.symbolOut = E.next.symbolOut;
   }
}

public static void FIELDREF(ast.FIELDREF E) {
   E.type = null /*Standard.NoType*/;
   E.next.typeOut = null /*Standard.NoType*/;
   E.next.symbolOut = null /*Standard.NoSy*/;

   if (! (E.typeIn instanceof sym.RecordType)) {
       aux.Error.Sem(E.position, "Record variable expected");
       E.type = E.typeOut = null /*Standard.NoType*/;
       E.symbol = null /*Standard.NoSy*/;
   } else {
      sym.RecordType t = (sym.RecordType) E.typeIn;
      sym.Symbol f = t.GetFields().locateByName(E.ident);
      E.symbol = f;
      if (f == null /*Standard.NoSy*/) {
         aux.Error.SemId(E.position, "Field identifier not declared", E.ident);
         E.type = E.typeOut = null /*Standard.NoType*/;
         E.symbol = null /*Standard.NoSy*/;
     } else {
	  sym.FieldSy field = (sym.FieldSy) f;
         E.next.env = E.env;
         E.next.typeIn = field.GetType();
         E.next.symbolIn = field;
         E.next.currentProc = E.currentProc;
         Designator(E.next);
         E.type = E.typeOut = E.next.typeOut;
         E.symbolOut = E.next.symbolOut;
     }
   }
}

public static void NARROW(ast.NARROW E) {
   E.type = null /*Standard.NoType*/;
   E.next.typeOut = null /*Standard.NoType*/;
   E.next.symbolOut = null /*Standard.NoSy*/;
   if (! (E.typeIn instanceof sym.ClassType))
     aux.Error.Sem(E.position, "Class type expected");

   E.next.typeIn = SemAux.ExpectType (E.ident, E.env, E.position);
   if (! (E.typeIn instanceof sym.ClassType))
      aux.Error.SemId(E.position, "Class type expected", E.ident);

   if (!SemAux.Assignable(E.typeIn, E.next.typeIn))
      aux.Error.Sem(E.position, "Illegal narrow.");

   E.next.env = E.env;
   E.next.symbolIn = E.symbolIn;
   E.next.currentProc = E.currentProc;
   Designator(E.next);

   E.type = E.typeOut = E.next.typeOut;
   E.symbolOut = E.next.symbolOut;
}

public static void CLASSREF(ast.CLASSREF E) {
   E.type = null /*Standard.NoType*/;
   E.next.typeOut = null /*Standard.NoType*/;
   E.next.symbolOut = null /*Standard.NoSy*/;

   if (! (E.typeIn instanceof sym.ClassType)) {
       aux.Error.Sem(E.position, "Class field expected");
       E.type = E.typeOut = null /*Standard.NoType*/;
       E.symbol = null /*Standard.NoSy*/;
   } else {
      sym.ClassType cl = (sym.ClassType) E.typeIn;
      E.symbol = SemAux.LocateClassField(cl, E.ident);
      if (E.symbol == null /*Standard.NoSy*/) {
         aux.Error.SemId(E.position, "Class field identifier not declared", E.ident);
         E.type = E.typeOut = null /*Standard.NoType*/;
         E.symbol = null /*Standard.NoSy*/;
     } else {
         E.next.env = E.env;
         if (E.symbol instanceof sym.MethodSy)
            E.next.typeIn = null /*Standard.NoType*/;
         else 
	     E.next.typeIn = ((sym.FieldSy)E.symbol).GetType();
         E.next.symbolIn = E.symbol;
         E.next.currentProc = E.currentProc;
         Designator(E.next);
         E.type = E.typeOut = E.next.typeOut;
         E.symbolOut = E.next.symbolOut;
     }
   }
}

public static void ENUMREF(ast.ENUMREF E) {
   E.type = null /*Standard.NoType*/;
   E.symbol = null;
   E.isLValue = false;
   E.isConst = true;
   sym.TypeSy ty = SemAux.ExpectType (E.typeName, E.env, E.position);
   if (ty == null) return;
   if (!(ty instanceof sym.EnumType))
       aux.Error.SemId(E.position, "Enumeration type expected", E.typeName);
   else {
      sym.EnumType t = (sym.EnumType) ty;
      E.type = t;
      sym.SyTab s = t.GetEnumElements();
      sym.EnumSy e = (sym.EnumSy) s.locateByName(E.idName);
      if (e == null /*Standard.NoSy*/)
          aux.Error.SemId(E.position, "Enumeration identifier not declared", E.idName);
      else 
         E.symbol = e;
   }
}

//===============================================================
//                         Expressions
//===============================================================
public static void Expression(ast.EXPRESSION E) {
        if (E instanceof ast.VARREF) Designator((ast.VARREF)E);
   else if (E instanceof ast.ENUMREF) Designator((ast.ENUMREF)E);
   else if (E instanceof ast.CHARLIT) CHARLIT((ast.CHARLIT)E);
   else if (E instanceof ast.STRINGLIT) STRINGLIT((ast.STRINGLIT)E);
   else if (E instanceof ast.INTLIT) INTLIT((ast.INTLIT)E);
   else if (E instanceof ast.REALLIT) REALLIT((ast.REALLIT)E);
   else if (E instanceof ast.BINARY) BINARY((ast.BINARY)E);
   else if (E instanceof ast.UNARY) UNARY((ast.UNARY)E);
   else if (E instanceof ast.NEW) NEW((ast.NEW)E);
   else if (E instanceof ast.ISA) ISA((ast.ISA)E);
   else aux.Error.Internal("Expression", "Node " +   E.getClass().getName() + " unexpected.");
}

public static void CHARLIT(ast.CHARLIT E) {
   E.type = sym.Standard.CharType;
   E.isLValue = false;
   E.isConst = true;
   E.value = new sym.CharValue(E.literal.charAt(0));
}

public static void STRINGLIT(ast.STRINGLIT E) {
   E.type = sym.Standard.StringType;
   E.isLValue = false;
   E.isConst = true;
   E.value = new sym.StringValue(E.literal);
}

public static void INTLIT(ast.INTLIT E) {
   E.type = sym.Standard.IntType;
   E.isConst = true;
   E.isLValue = false;
   try {
      E.value = new sym.IntValue(java.lang.Integer.parseInt(E.literal));
   } catch (java.lang.NumberFormatException e) {
       aux.Error.Sem(E.position, "Illegal integer constant");
       E.value = new sym.IntValue(0);
   }
}

public static void REALLIT(ast.REALLIT E) {
   E.type = sym.Standard.RealType;
   E.isConst = true;
   E.isLValue = false;
   E.value = new sym.RealValue(java.lang.Float.parseFloat(E.literal));
}

public static void ISA(ast.ISA E) {
   E.isLValue = false;
   E.isConst = false;
   E.value = null;
   //   E.symbol = E.type = SemAux.ExpectType (E.typeName, E.env, E.position);
   sym.TypeSy R = SemAux.ExpectTypeKind (E.typeName, E.env, sym.ClassType.class, "class", E.position);

   E.left.env = E.env;
   E.left.currentProc = E.currentProc;
   Expression(E.left);

   if (!(E.left.type instanceof sym.ClassType)) {
       aux.Error.Sem(E.position, "Class type expected");
   } else {
       if (R != null) {
           sym.ClassType Right = (sym.ClassType)R;
           sym.ClassType Left = (sym.ClassType) E.left.type;
           if (! (SemAux.IsSuperType(Right,Left) || SemAux.IsSuperType(Left,Right)))
              aux.Error.Sem(E.position, "Type mismatch");
       }
   }
   E.type = sym.Standard.BoolType;
}

public static void NEW(ast.NEW E) {
   E.isLValue = false;
   E.isConst = false;
   E.value = null;
   E.symbol = E.type = SemAux.ExpectType (E.typeName, E.env, E.position);
   if (E.type == null) return;
   if ((! (E.type instanceof sym.RefType)) &&
       (! (E.type instanceof sym.ClassType)))
          aux.Error.SemId(E.position, "Pointer or class type expected", E.typeName);
}


public static void BINARY(ast.BINARY E) {
   E.isLValue = false;
   E.left.env = E.right.env = E.env;
   switch (E.op) {
   case lexer.Token.PLUS:
   case lexer.Token.MINUS:
   case lexer.Token.STAR:
   case lexer.Token.SLASH:
          E.left.currentProc = E.currentProc;
          Expression(E.left);
          E.right.currentProc = E.currentProc;
          Expression(E.right);
          E.left.type  = SemAux.CheckNumeric(E.left.type, E.position);
          E.right.type = SemAux.CheckNumeric(E.right.type, E.position);
          E.type = SemAux.CheckBinary(E.left.type, E.right.type, E.left.type, E.position);
          break;
   case lexer.Token.PERCENT:
          E.left.currentProc = E.currentProc;
          Expression(E.left);
          E.right.currentProc = E.currentProc;
          Expression(E.right);
          E.left.type  = SemAux.CheckInteger(E.left.type, E.position);
          E.right.type = SemAux.CheckInteger(E.right.type, E.position);
          E.type = SemAux.CheckBinary(E.left.type, E.right.type, E.left.type, E.position);
          break;
   case lexer.Token.EQ:
   case lexer.Token.NE:
          E.left.currentProc = E.currentProc;
          Expression(E.left);
          E.right.currentProc = E.currentProc;
          Expression(E.right);
          E.left.type  = SemAux.CheckComparable(E.left.type, E.position);
          E.right.type = SemAux.CheckComparable(E.right.type, E.position);
          E.type = SemAux.CheckBinary(E.left.type, E.right.type, sym.Standard.BoolType, E.position);
          break;
   case lexer.Token.LT:
   case lexer.Token.GT:
   case lexer.Token.LE:
   case lexer.Token.GE:
          E.left.currentProc = E.currentProc;
          Expression(E.left);
          E.right.currentProc = E.currentProc;
          Expression(E.right);
          E.left.type  = SemAux.CheckScalar(E.left.type, E.position);
          E.right.type = SemAux.CheckScalar(E.right.type, E.position);
          E.type = SemAux.CheckBinary(E.left.type, E.right.type, sym.Standard.BoolType, E.position);
          break;
   case lexer.Token.AND:
          E.left.false_ = E.false_;
          E.left.true_ = new sym.Label();
          E.right.true_ = E.true_;
          E.right.false_ = E.false_;
          E.left.currentProc = E.currentProc;
          Expression(E.left);
          E.right.currentProc = E.currentProc;
          Expression(E.right);
          E.type = SemAux.CheckBinaryBool(E.left.type, E.right.type, E.position);
          break;
   case lexer.Token.OR:
          E.left.true_ = E.true_;
          E.left.false_ = new sym.Label();
          E.right.true_ = E.true_;
          E.right.false_ = E.false_;
          E.left.currentProc = E.currentProc;
          Expression(E.left);
          E.right.currentProc = E.currentProc;
          Expression(E.right);
          E.type = SemAux.CheckBinaryBool(E.left.type, E.right.type, E.position);
   }
   E.isConst = E.left.isConst && E.right.isConst;
   if (E.isConst && E.left.type != null && E.right.type != null && E.type != null)
      E.value = SemAux.EvalBinary(E.isConst, E.left.value, E.right.value, E.left.type, E.op, E.position);
}

public static void UNARY(ast.UNARY E) {
   E.left.env = E.env;
   switch (E.op) {
   case lexer.Token.MINUS:
          E.left.currentProc = E.currentProc;
          Expression(E.left);
          E.type  = SemAux.CheckNumeric(E.left.type, E.position);
          break;
   case lexer.Token.NOT:
         E.left.true_ = E.false_;
         E.left.false_ = E.true_;
         E.left.currentProc = E.currentProc;
         Expression(E.left);
         E.type = SemAux.CheckBool(E.left.type, E.position);
         break;
   case lexer.Token.TRUNC:
         E.left.currentProc = E.currentProc;
         Expression(E.left);
         if (E.left.type != sym.Standard.RealType) {
            aux.Error.Sem(E.position, "Real type expected");
            E.type = null /*Standard.NoType*/;
         } else
            E.type = sym.Standard.IntType;
         break;
   case lexer.Token.FLOAT:
         E.left.currentProc = E.currentProc;
         Expression(E.left);
         if (E.left.type != sym.Standard.IntType) {
            aux.Error.Sem(E.position, "Integer type expected");
            E.type = null /*Standard.NoType*/;
         } else
            E.type = sym.Standard.RealType;
      }
      E.isConst = E.left.isConst;
      if (E.isConst && E.left.type != null && E.type != null)
         E.value = SemAux.EvalUnary(E.isConst, E.left.value, E.left.type, E.op, E.position);
}
}   
